@@ -87,6 +87,8 @@ parser.add_argument("interval", nargs="?", default=-1,
     help="output interval, in seconds")
 parser.add_argument("count", nargs="?", default=99999999,
     help="number of outputs")
+parser.add_argument("--ebpf", action="store_true",
+    help=argparse.SUPPRESS)
 args = parser.parse_args()
 countdown = int(args.count)
 frequency = 99
@@ -155,8 +157,10 @@ int do_perf_event(struct bpf_perf_event_data *ctx)
 """
 
 # code substitutions
-if debug:
+if debug or args.ebpf:
     print(bpf_text)
+    if args.ebpf:
+        exit()
 
 # initialize BPF & perf_events
 b = BPF(text=bpf_text)
@@ -214,7 +218,7 @@ while 1:
             sleep(wakeup_s)
     except KeyboardInterrupt:
         exiting = 1
-    b.kprobe_poll()
+    b.perf_buffer_poll()
     slept += wakeup_s
 
     if slept < 0.999 * interval:   # floating point workaround
