@@ -32,6 +32,8 @@ parser.add_argument("-N", "--netns", default=0, type=int,
                     help="trace this Network Namespace only")
 parser.add_argument("-v", "--verbose", action="store_true",
                     help="include Network Namespace in the output")
+parser.add_argument("--ebpf", action="store_true",
+                    help=ap.SUPPRESS)
 args = parser.parse_args()
 
 bpf_text = """
@@ -433,7 +435,7 @@ int trace_accept_return(struct pt_regs *ctx)
 
   // Get network namespace id, if kernel supports it
 #ifdef CONFIG_NET_NS
-  possible_net_t skc_net;
+  possible_net_t skc_net = { };
   bpf_probe_read(&skc_net, sizeof(skc_net), &newsk->__sk_common.skc_net);
   bpf_probe_read(&net_ns_inum, sizeof(net_ns_inum), &skc_net.net->ns.inum);
 #else
@@ -622,6 +624,10 @@ if args.netns:
 bpf_text = bpf_text.replace('##FILTER_PID##', pid_filter)
 bpf_text = bpf_text.replace('##FILTER_NETNS##', netns_filter)
 
+if args.ebpf:
+    print(bpf_text)
+    exit()
+
 # initialize BPF
 b = BPF(text=bpf_text)
 b.attach_kprobe(event="tcp_v4_connect", fn_name="trace_connect_v4_entry")
@@ -664,4 +670,4 @@ def inet_ntoa(addr):
 b["tcp_ipv4_event"].open_perf_buffer(print_ipv4_event)
 b["tcp_ipv6_event"].open_perf_buffer(print_ipv6_event)
 while True:
-    b.kprobe_poll()
+    b.perf_buffer_poll()
